/**
 * This file contains the logic of the NavigationView component.
 * It is used to browse through the structure of a project, represented in a JSON file.
 *
 * @author baltes
 * @see NavigationView.html, NavigationView.css
 *
 */
var NavigationView = (function(util, icons) {

    // TODO : use hashes for path testing? better performance?

    var NavigationView = function() {
        //region Constants
        var INITIAL_PADDING = 12;
        var INDENTION_STEP = 20;
        var VERTICAL_ENTITY_SEPARATION = 12;
        var HEADER_HEIGHT = 56;
        var ELEMENT_COLOR_SELECTED = '#558FED';
        var BORDER_COLOR = '#555555';
        var BORDER_COLOR_SELECTED = '#417BD9';
        //endregion

        //region Private members
        var that,
            filterString,
            /* color scales */
            moduleColorScale,
            selectedModuleColorScale,
            entityColorScale,
            selectedEntityColorScale,
            /* containers */
            availableHeight,
            leftOuterContainer,
            moduleContainer,
            rightOuterContainer,
            entityContainer,
            /* variables for d3.js */
            nodes, modules, entities, // arrays with nodes
            moduleDivs, entityDivs, // selections with module divs
            projectNode, selectedModuleNode, selectedEntityNode, // single DOM nodes
            /* GUI elements */
            projectSelector,
            $projectSelector,
            filterTextInputNode,
            foldedSliderNode,
            /* data exchange with mainpage */
            artifactPath,
            artifactType,
            defaultArtifactPath,
            defaultArtifactType;

        initialize();
        //endregion

        //region Getter/setter for private members
        Object.defineProperties(this, {
            "defaultArtifactPath": {
                get: function() {
                    return defaultArtifactPath;
                },
                set: function(newValue) {
                    defaultArtifactPath = newValue;
                }
            },
            "defaultArtifactType": {
                get: function() {
                    return defaultArtifactType;
                },
                set: function(newValue) {
                    defaultArtifactType = newValue;
                }
            },
            "artifactPath": {
                get: function() {
                    return artifactPath;
                },
                set: function(newValue) {
                    artifactPath = newValue;
                }
            },
            "artifactType": {
                get: function() {
                    return artifactType;
                },
                set: function(newValue) {
                    artifactType = newValue;
                }
            }
        });
        //endregion

        //region Private methods
        function initialize() {
            icons.load();

            that = this;
            filterString = "";

            /* color scales */
            moduleColorScale = d3.scale.linear()
                .domain([0, 6])
                .range(["#969696", "#EAEAEA"]);
            selectedModuleColorScale = d3.scale.linear()
                .domain([0, 6])
                .range(["#7DADFF", "#E5EFFF"]);
            entityColorScale = d3.scale.linear()
                .domain([0, 1])
                .range(["#E0E0E0", "#EAEAEA"]);
            selectedEntityColorScale = d3.scale.linear()
                .domain([0, 1])
                .range(["#5F99F7", "#8CC1FF"]);

            /* configure containers */
            availableHeight = window.innerHeight;
            /* left */
            leftOuterContainer = d3.select("#leftOuterContainer");
            moduleContainer = d3.select("#moduleContainer");
            /* right*/
            rightOuterContainer = d3.select("#rightOuterContainer");
            entityContainer = d3.select("#entityContainer");

            resize();

            /* variables for D3.js */
            projectSelector = d3.select('#projectSelector');
            $projectSelector = $('#projectSelector');

            filterTextInputNode = d3.select("#filterTextInput")
                .on("keypress", updateFilterString)
                .on("keyup", updateFilterString)
                .on("paste", updateFilterString)
                .node();

            foldedSliderNode = d3.select("#foldedSlider")
                .node();

            /* register event handlers */
            d3.select(window)
                .on("resize", resize);
            $("#foldedSliderDiv")
                .on("change", foldSliderChange);
            d3.select('.header-b.middle > div > a') // clear button in search input
                .on("click", function() {
                    util.log("clear_filter_button_clicked");
                    updateFilterString();
                });
            projectSelector
                .on("change", projectChange);
        }

        function loadProjectIndex(pathToIndexFile, callback) {
            var i,
                projectName;

            d3.json(pathToIndexFile, function(jsonArray) {
                /* update project selector */
                projectSelector
                    .selectAll("option")
                    .remove();
                for (i=0; i<jsonArray.length; i++) {
                    projectSelector
                        .append("option")
                        .attr("value", function() {
                            return jsonArray[i].name;
                        })
                        .text(function() {
                            return jsonArray[i].name;
                        })
                }
                $projectSelector.selectmenu("refresh", true);

                if (artifactPath && artifactType) { // artifactPath and artifactType for initial selection are set
                    selectArtifact(artifactPath, artifactType, callback);
                } else if (defaultArtifactPath && defaultArtifactType) { // default values are set (=> whole sketch is linked)
                    selectArtifact(defaultArtifactPath, defaultArtifactType, callback);
                } else {
                    /* new selection, load first project */
                    selectedModuleNode = null;
                    selectedEntityNode = null;

                    if (projectSelector.property("options")[0]) {
                        projectName = projectSelector.property("options")[0].value;
                        selectProject(projectName);

                        loadProject(projectSelector.property("options")[0].value, function() {
                            highlightArtifact();
                            scrollTop();

                            /* execute callback function */
                            if (callback) {
                                callback();
                            }
                        });
                    }
                }
            });
        }

        function selectArtifact(newArtifactPath, newArtifactType, callback) {
            var path,
                node,
                moduleNode,
                projectName,
                projectOptionsIndex;

            selectedModuleNode = null;
            selectedEntityNode = null;

            if (newArtifactPath && newArtifactType) {
                // load project for selected artifact
                path = util.tokenizePathString(newArtifactPath);

                projectName = path[0];

                loadProject(projectName, function() {
                    node = util.validatePathArrayWithNode(projectNode, path, 0);

                    if (!node) {
                        throw new ReferenceError(APPLICATION_NAME + ": Artifact not found: " + newArtifactPath);
                    }

                    if (node.elementType != newArtifactType) {
                        throw new ReferenceError(APPLICATION_NAME + ": Artifact type should be " + newArtifactType + ", but it is " + node.elementType);
                    }

                    selectProject(projectName);

                    if (newArtifactType == "project" || newArtifactType == "module") {
                        selectedModuleNode = node;
                        highlightModule(selectedModuleNode, true);
                        scrollModuleIntoView(node);
                    } else if (newArtifactType == "entity" || newArtifactType == "subentity") {
                        selectedEntityNode = node;
                        highlightEntity(selectedEntityNode, true);
                        selectedModuleNode = getContainingModule(selectedEntityNode);
                        highlightModule(selectedModuleNode, true);
                        updateEntityContainer(selectedModuleNode);
                    }

                    scrollSelectedArtifactIntoView();

                    artifactPath = newArtifactPath;
                    artifactType = newArtifactType;

                    /* execute callback function */
                    if (callback) {
                        callback();
                    }
                });
            }
        }

        function selectProject(projectName) {
            var projectOptionsIndex;

            projectOptionsIndex = 0;
            $projectSelector.find("option").each(
                function(index) {
                    if ($(this).text() == projectName) {
                        projectOptionsIndex = index;
                    }
                }
            );
            /* set current project in selector */
            projectSelector.property("selectedIndex", projectOptionsIndex);
            $projectSelector.selectmenu("refresh");

            return projectOptionsIndex;
        }

        function scrollTop() {
            moduleContainer
                .property("scrollTop", 0)
                .property("scrollLeft", 0);
            entityContainer
                .property("scrollTop", 0)
                .property("scrollLeft", 0)
        }

        function loadProject(projectName, callback) {
            /* load JSON file with project structure into containers */
            d3.json("/projects/" + projectName, function(jsonObject) {
                var treeLayout;

                if (moduleDivs) {
                    moduleDivs.remove();
                }

                if (entityDivs) {
                    entityDivs.remove();
                }

                treeLayout = d3.layout.tree();

                /* create nodes using tree layout */
                nodes = treeLayout.nodes(jsonObject);

                /* filter nodes for module container */
                modules = nodes.filter( function(node) {
                    return (node.elementType == "project" || node.elementType == "module");
                });

                /* filter nodes for entity container */
                entities = nodes.filter( function(node) {
                    return ((node.elementType == "entity") || (node.elementType == "subentity"));
                });

                /* create module divs in module container */
                moduleDivs = moduleContainer
                    .selectAll("div.element")
                    .data(modules)
                    .enter()
                    .append("div")
                    .text(function(d) {
                        return d.elementName;
                    })
                    .classed("element", true)
                    .classed("icon-project-nav bold-font", function(d) {
                        return (d.elementType == "project");
                    })
                    .classed("icon-module-nav", function(d) {
                        return (d.elementType == "module");
                    })
                    .style("margin-left", function(d) {
                        return INITIAL_PADDING + d.depth*INDENTION_STEP + "px";
                    })
                    .style("background-color", function(d) {
                        return moduleColorScale(d.depth);
                    })
                    .on("click", moduleClick);

                /* set top margin for first module div */
                moduleContainer
                    .select("div:first-child")
                    .style("margin-top", "12px");

                /* project node is root */
                projectNode = modules[0];

                /* create entity divs in entity container (match all elements and don't hide subentities)*/
                updateEntityContainer(projectNode, function() {
                    return true;
                }, false);

                if (callback) {
                    callback();
                }
            });
        }

        function projectChange() {
            var selectedProjectIndex,
                projectName;

            parent.$.mobile.loading('show');

            selectedProjectIndex = projectSelector.property("selectedIndex");

            if (selectedProjectIndex != -1) {
                projectName = projectSelector.property("options")[selectedProjectIndex].value;

                util.log("change_project;" + projectName);

                loadProject(projectName, function() {
                    highlightArtifact();
                    parent.$.mobile.loading('hide');
                });
            }
        }

        function scrollSelectedArtifactIntoView() {
            scrollModuleIntoView(selectedModuleNode);
            scrollEntityIntoView(selectedEntityNode);
        }

        function scrollModuleIntoView(moduleNode) {
            var moduleDiv;

            if (moduleNode) {
                moduleDiv = moduleDivs.filter(function(d) { return (d == moduleNode) }).node();
                moduleDiv.scrollIntoView(true);
            }
        }

        function scrollEntityIntoView(entityNode) {
            var entityDiv;

            if (entityNode) {
                entityDiv = entityDivs.filter(function(d) { return (d == entityNode) }).node();
                entityDiv.scrollIntoView(true);
            }
        }

        function resetColor() {
            if (moduleDivs) {
                /* reset color of module divs */
                setDivColor(moduleDivs, BORDER_COLOR, function(d) {
                    return moduleColorScale(d.depth);
                });
            }

            if (entityDivs) {
                /* reset color of entity divs */
                setDivColor(entityDivs, BORDER_COLOR, function(d) {
                    return entityColorScale(getEntityDepth(d));
                });
            }
        }

        function highlightArtifact() {
            var path,
                selectedNode;

            resetColor();

            if (artifactPath && artifactType) { // highlight selected artifact
                path = util.tokenizePathString(artifactPath);
                selectedNode = util.validatePathArrayWithNode(projectNode, path, 0);

                if (!selectedNode || selectedNode.elementType != artifactType) {
                    return; // artifact does not exist in current project
                }

                if (selectedNode.elementType == "project" || selectedNode.elementType == "module") {
                    selectedModuleNode = selectedNode;
                    highlightModule(selectedModuleNode, true);
                } else if (selectedNode.elementType == "entity" || selectedNode.elementType == "subentity") {
                    selectedEntityNode = selectedNode;
                    highlightEntity(selectedEntityNode, true);
                    selectedModuleNode = getContainingModule(selectedEntityNode);
                    highlightModule(selectedModuleNode, true);
                }
            }
        }

        function resize() {
            var availableHeight = window.innerHeight-2;

            /* left */
            leftOuterContainer
                .style("height", availableHeight + "px");
            moduleContainer
                .style("height", availableHeight-HEADER_HEIGHT + "px");

            /* right */
            rightOuterContainer
                .style("height", availableHeight + "px");
            entityContainer
                .style("height", availableHeight-HEADER_HEIGHT + "px");
        }

        function foldSliderChange() {
            filterEntities(foldedSliderNode.value == "folded");

            util.log("change_folderslider;" + foldedSliderNode.value);
        }

        function updateFilterString() {
            var newFilterString,
                newCharacter;

            newFilterString = filterTextInputNode.value;

            if (d3.event.type == "keypress") {
                newCharacter = String.fromCharCode(Number(d3.event.which));

                if (/[a-zA-Z0-9-_]/.test(newCharacter)) {
                    newFilterString += newCharacter;
                } else if (newCharacter == '\r') { // prevent default action when return is pressed, since pressing this key broke the page
                    d3.event.preventDefault();
                }
            }

            if (newFilterString != filterString) {
                filterString = newFilterString;
                filterEntities(foldedSliderNode.value == "folded");
                util.log("change_filter;" + filterString);
            }
        }

        function filterEntities(hideSubentities) {
            updateEntityContainer(selectedModuleNode, function filter(node) {
                return (node.elementName.toLowerCase().indexOf(filterString.toLowerCase()) != -1);
            }, hideSubentities);
        }

        function moduleClick(clickedNode) {
            selectedModuleNode = clickedNode;
            selectedEntityNode = null;

            highlightModule(clickedNode, true);

            updateEntityContainer(
                clickedNode,
                function () { return true; },
                false
            );

            filterEntities(foldedSliderNode.value == "folded");

            saveSelectedArtifact();

            util.log("module_clicked;" + clickedNode.elementName);
        }

        function highlightModule(moduleNode, highlightChildren) {
            /* reset color of module divs */
            setDivColor(moduleDivs, BORDER_COLOR, function(d) {
                return moduleColorScale(d.depth);
            });

            /* highlight selected module div */
            setDivColor(
                moduleDivs.filter( function(d) { return (d == moduleNode); }),
                BORDER_COLOR_SELECTED,
                function () { return ELEMENT_COLOR_SELECTED; }
            );

            if (highlightChildren) {
                highlightChildDivs(moduleDivs, moduleNode, function(d) {
                    return selectedModuleColorScale(d.depth);
                });
            }
        }

        function entityClick(clickedNode) {
            var selectedModuleDiv;

            selectedEntityNode = clickedNode;
            selectedModuleNode = getContainingModule(clickedNode);

            highlightModule(selectedModuleNode, false);
            highlightEntity(selectedEntityNode, true);

            /* scroll selected module into view if not visible */
            selectedModuleDiv = moduleDivs.filter(function(d) { return (d == selectedModuleNode) }).node();
            if (!util.isScrolledIntoView(selectedModuleDiv, entityContainer.node())) {
                scrollModuleIntoView(selectedModuleNode);
            }

            saveSelectedArtifact();

            util.log("entity_clicked;" + clickedNode.elementName);
        }

        function getContainingModule(node) {
            var moduleNode;

            switch (node.elementType) {
                case "entity":
                    moduleNode = node.parent;
                    break;
                case "subentity":
                    moduleNode = node.parent.parent;
                    break;
                case "module":
                    moduleNode = node.parent;
                    break;
                default:
                    moduleNode = null;
            }

            return moduleNode;
        }

        function highlightEntity(entityNode, highlightChildren) {
            /* reset color of entity divs */
            setDivColor(entityDivs, BORDER_COLOR, function(d) {
                return entityColorScale(getEntityDepth(d));
            });

            /* highlight clicked entity div */
            setDivColor(
                entityDivs.filter( function(d) { return (d == entityNode); }),
                BORDER_COLOR_SELECTED,
                function () { return ELEMENT_COLOR_SELECTED; }
            );

            if (highlightChildren) {
                highlightChildDivs(entityDivs, entityNode, function(d) {
                    return selectedEntityColorScale(getEntityDepth(d));
                });
            }
        }

        function getEntityDepth(node) {

            var depth;

            switch (node.elementType) {
                case "entity":      depth = 0;
                    break;
                case "subentity":   depth = 1;
                    break;
                default:            depth = -1;
            }

            return depth;
        }

        function getAllChildEntities(node, filter) {
            var childNodes = [];
            var recursiveChildNodes = [];
            var currentNode;
            var i, j;

            if (node.children) {

                for (i=0; i<node.children.length; i++) {
                    currentNode = node.children[i];

                    if (currentNode.elementType == "entity") {
                        if (filter(currentNode)) { // filter matches entity
                            childNodes.push(currentNode);
                            if (currentNode.children) { // maybe the entity has no subentities
                                childNodes = childNodes.concat(currentNode.children); // add the subentities also to the result if the entity has been matched
                            }
                        }
                    }
                    else if (currentNode.elementType == "subentity") {
                        if (filter(currentNode)) {
                            if (childNodes.indexOf(currentNode.parent) == -1) {
                                // only add parent if not already in childNodes
                                childNodes.push(currentNode.parent);
                            }
                            childNodes.push(currentNode);
                        }
                    }

                    if (currentNode.children) { // currentNode is module or entity with children
                        recursiveChildNodes = getAllChildEntities(currentNode, filter);

                        for (j=0; j<recursiveChildNodes.length; j++) {
                            currentNode = recursiveChildNodes[j];
                            if (childNodes.indexOf(currentNode) == -1) {
                                // only add node if not already in childNodes
                                childNodes.push(currentNode);
                            }
                        }
                    }
                }
            }

            return childNodes;
        }

        function createEntityDivs(selection, data) {
            return selection
                .selectAll("div.element")
                .data(data)
                .enter()
                .append("div")
                .classed("element", true)
                .classed("icon-entity-nav", function(d) {
                    return (d.elementType == "entity");
                })
                .classed("icon-subentity-nav", function(d) {
                    return (d.elementType == "subentity");
                })
                .style("margin-left", function(d) {
                    return INITIAL_PADDING + getEntityDepth(d)*INDENTION_STEP + "px";
                })
                .style("background-color", function(d) {
                    return entityColorScale(getEntityDepth(d));
                })
                .text(function(d) {
                    return d.elementName;
                })
                .on("click", entityClick);
        }

        function setDivColor(selection, borderColor, backgroundColorFunction) {
            selection
                .style("border-color", borderColor)
                .style("background-color", function(d) {
                    return backgroundColorFunction(d);
                });
        }

        function highlightChildDivs(selection, node, colorFunction) {
            var i;

            /* filter selection for child divs and set highlight color */
            setDivColor(
                selection.filter(function(d) { return (d.parent == node) }),
                BORDER_COLOR_SELECTED,
                colorFunction
            );

            /* visit child nodes */
            if (node.children) {
                for (i=0; i<node.children.length; i++) {
                    highlightChildDivs(selection, node.children[i], colorFunction);
                }
            }
        }

        function updateEntityContainer(moduleNode, filter, hideSubentities) {
            var childEntities,
                filterFunction;

            if (!entityContainer) {
                return;
            }

            if (filter) {
                filterFunction = filter;
            } else {
                filterFunction = function() {
                    return true;
                }
            }

            /* if no module is selected (e.g. on startup), use the project node */
            if (moduleNode) {
                childEntities = getAllChildEntities(moduleNode, filterFunction);
            } else if (projectNode) {
                childEntities = getAllChildEntities(projectNode, filterFunction);
            } else {
                return;
            }

            /* scroll to top position and clear entity container */
            entityContainer
                .property("scrollTop", 0)
                .property("scrollLeft", 0)
                .selectAll("div.element").remove();

            if (hideSubentities) {
                childEntities = childEntities.filter(function (node) {
                    return (node.elementType != "subentity");
                });
            }

            /* append divs for child nodes to entity container */
            if (childEntities.length > 0) {
                entityDivs = createEntityDivs(entityContainer, childEntities);
            }

            /* separate entities vertically from each other */
            entityContainer.selectAll("div.icon-entity-nav")
                .style("margin-top", VERTICAL_ENTITY_SEPARATION + "px");

            /* separate the last div from bottom of container */
            entityContainer.selectAll("div:last-child")
                .style("margin-bottom", VERTICAL_ENTITY_SEPARATION + "px");

            /* if an entity is selected then highlight it */
            if (selectedEntityNode) {
                highlightEntity(selectedEntityNode, true);
            }
        }

        function saveSelectedArtifact() {
            if (selectedEntityNode) { // entity
                artifactPath =  util.getPathStringFromNode(selectedEntityNode);
                artifactType = selectedEntityNode.elementType;
            } else if (selectedModuleNode) { // module or project
                artifactPath =  util.getPathStringFromNode(selectedModuleNode);
                artifactType = selectedModuleNode.elementType;
            }
        }

        function clear() {
            var $foldSlider,
                $filterTextInput;

            util.disableLogging();

            $foldSlider = $(foldedSliderNode);
            $filterTextInput = $(filterTextInputNode);

            resetColor();
            selectedModuleNode = null;
            selectedEntityNode = null;
            artifactPath = "";
            artifactType = "";
            defaultProject = "";

            filterString = "";
            $filterTextInput.val('');
            $filterTextInput.change();

            $foldSlider.val("unfolded");
            $foldSlider.slider("refresh");
            foldSliderChange();

            util.enableLogging();
        }
        //endregion

        //region Privileged methods
        this.loadProjectIndex = loadProjectIndex;
        this.clear = clear;
        //endregion
    };

    return NavigationView;

})(util, icons);

var NavigationPage = (function(NavigationView) {
    $(document).on("pageinit", "#navigationPage", function() {
        window.navigationView = new NavigationView();
    });
})(NavigationView);