/**
 * Main component for SVG canvas, manages drawing, modifying and deletion of rectanges, opening of panel and metainfo popups, file upload, etc.
 *
 * @author baltes
 * @see SvgSelection.js, SvgViewbox.js
 *
 */
var SvgCanvas = (function(util, PanelPopup, MetaInfoPopup, NavigationPopup, SvgSelection) {

    var SvgCanvas = function(args) {
        //region Constants
        Object.defineProperties(this, {
            "MIN_LENGTH": {
                value: 20
            },
            "SIZE_THRESHOLD": {
                value: 5 // needed for resize event, problems with iOS 6.1
            },
            "OUTLINE_RECT_STYLE": {
                value: "stroke: #0000ff; stroke-width: 2; fill: none; opacity: 0.5;"
            },
            "RECT_STYLE": {
                value: "stroke: #0000ff; stroke-width: 2; fill: #0000ff; fill-opacity: 0.08;"
            },
            "CIRCLE_STYLE": {
                value: "stroke: #d8d8d8; stroke-width: 2; fill: #0000ff; fill-opacity: 0.6;"
            }
        });
        //endregion

        //region Private members
        var that,
            svgURI,
            pathToProjectIndex,
            page,
            $page,
            div,
            $div,
            root,
            $root,
            width,
            height,
            divOffset,
            isMouseDown,
            startPosCursor,
            history,
            historyIndex,
            outlineRect,
            currentInsertionPos,
            visibleArea,
            disableRectClick,
            disableCanvasClick,
            viewBox,
            selection,
            setTitleEvent,
            metaData,
            $image,
            initialWidth,
            initialHeight;

        that = this;
        page = args.page;
        $page = $(page);
        div = args.svgDiv;
        $div = $(div);
        selection = new SvgSelection({
            svgCanvas: that
        });

        /* custom event to trigger setting header title of main page */
        setTitleEvent = new CustomEvent(
            "setTitle",
            {
                detail: {
                    text: "" // new title for main page
                },
                bubbles: true,
                cancelable: true
            }
        );

        /* register event handlers */
        registerGlobalEventHandlers();
        registerSvgEventHandlers();

        /* create popups and store reference in global popups object */
        popups.panelPopup = new PanelPopup({
            svgCanvas: that
        });
        popups.metaInfoPopup = new MetaInfoPopup({
            svgCanvas: that
        });

        popups.panelPopup.create(function() {
            popups.metaInfoPopup.create();
        });

        initialize();
        //endregion

        //region Getter/setter for private members
        Object.defineProperties(this, {
            "page": {
                get: function() {
                    return page;
                }
            },
            "$page": {
                get: function() {
                    return $page;
                }
            },
            "root": {
                get: function() {
                    return root;
                }
            },
            "$root": {
                get: function() {
                    return $root;
                }
            },
            "isMouseDown": {
                get: function() {
                    return isMouseDown;
                }
            }
        });
        //endregion

        //region Private methods
        function initialize(svgURI) {
            $div.addClass("sl-hidden"); // canvas div is initially hidden and is shown after initialization

            $div.empty(); // clear canvas

            divOffset = 0;
            isMouseDown = false;
            startPosCursor = null;
            root = null;
            $root = null;
            history = [];
            historyIndex = 0;
            outlineRect = null;
            currentInsertionPos = null;
            visibleArea = null;
            disableRectClick = false;
            disableCanvasClick = false;

            if (svgURI) {
                loadSvg(svgURI);

                // get sketch metadata
                $.get("/sketches/metadata/"+$root.attr("id"), function (data) {
                    setTitleEvent.detail.text = data.title;
                    $page.get(0).dispatchEvent(setTitleEvent);
                });

                registerSvgEventHandlers();
                resize();
            } else {
                // trigger event to hide header title of main page
                setTitleEvent.detail.text = '';
                $page.get(0).dispatchEvent(setTitleEvent);

            }

            $div.removeClass("sl-hidden");
        }

        function loadSvg(svgURI) {
            $.ajax({
                url: svgURI,
                dataType: 'xml',
                async: false,
                success: function (svgDoc) {
                    /* get root element of the svg document */
                    var loadedSvgRoot = svgDoc.documentElement;
                    /* make a deep copy of the svg root node that can be inserted into the current document */
                    root = document.importNode(loadedSvgRoot, true);
                    $root = $(root);

                    //$image = $root.find('image:first');
                    //initialWidth = parseInt($image.attr("width").replace("px", ""));
                    //initialHeight = parseInt($image.attr("height").replace("px", ""));

                    /* append the imported svg root element to the div */
                    $div.append(root);

                    /* clear changelog */
                    window.changelog = {};
                },
                error: function(jqXHR) {
                    console.log("An ajax error occured.\n\n"
                        + "Status: " + jqXHR.status + "\n"
                        + "Status text: " + jqXHR.statusText);
                    $.mobile.loading('hide');
                }
            });
        }

        function uploadSvg() {
            var $mainGroup,
                id,
                data,
                xmlSerializer,
                svgUrl;

            if (!$root) {
                return;
            }

            /* remove circles and close panel */
            selection.clear();
            if (popups.panelPopup.isOpen) {
                popups.panelPopup.close();
            }

            id = $root.attr('id');

            /* Don't upload attributes for SVG root */
            var svgWidth = $root.attr('width');
            var svgHeight = $root.attr('height');
            var svgStyle = $root.attr('style');
            $root.removeAttr('width');
            $root.removeAttr('height');
            $root.removeAttr('style');

            xmlSerializer = new XMLSerializer();

            data = {
                id: id,
                svg: xmlSerializer.serializeToString(root)
            };

            if (window.changelog) {
                data["changelog"] = JSON.stringify(window.changelog);
            }

            svgUrl = "/sketches/svg/" + id;

            $.ajax({
                type: "POST",
                url: svgUrl,
                data: data,
                dataType: "xml",
                processData: "false",
                success: function(jsonData) {
                    alert("Upload successful!");
                    //util.log("upload_changes;" + svgUrl);
                    window.changelog = {};
                    window.code_id = "";
                },
                error: function(jqXHR) {
                    alert("An ajax error occured.\n\n"
                        + "Status: " + jqXHR.status + "\n"
                        + "Status text: " + jqXHR.statusText);
                    $.mobile.loading('hide');
                }
            });

            if (svgWidth) {
                $root.attr('width', svgWidth);
            }
            if (svgHeight) {
                $root.attr('height', svgHeight);
            }
            if (svgStyle) {
                $root.attr('style', svgStyle);
            }
        }

        function registerGlobalEventHandlers() {
            util.disableTextSelection($div);

            $(document).on('vmouseup', '*', that, function () {
                isMouseDown = false;
            });

            $(window).on('resize', function(event) {
                if (!(util.isIOSDevice() && util.isInputSelected())) {
                    resize();
                }
            });
        }

        function registerSvgEventHandlers() {
            // register event handlers for drawing of rectangles
            util.addEventListener(root, 'vmousedown', mouseDown, that);
            util.addEventListener(root, 'vmousemove', mouseMove, that);
            util.addEventListener(root, 'vmouseup', mouseUp, that);
            util.addEventListener(root, 'vclick', click, that);
        }

        function resize() {
            var newVisibleArea;

            newVisibleArea = util.getVisibleArea();

            if (visibleArea) {
                if (( Math.abs(visibleArea.height - newVisibleArea.height) <= that.SIZE_THRESHOLD)
                        && (Math.abs(visibleArea.width - newVisibleArea.width) <= that.SIZE_THRESHOLD)) {
                    return;
                }
            }

            visibleArea = newVisibleArea;

            width = newVisibleArea.width;
            height = newVisibleArea.height;

            $div
                .css("width", width + "px")
                .css("height", height + "px");

            if ($root) {
                setSvgAttributes({
                    width: width + "px",
                    height: height + "px",
                    style: "width: " +  width + "px; height: " +  height + "px;"
                });
            }
        }

        function mouseDown(event) {

            event.preventDefault();

            startPosCursor = {
                x: event.clientX,
                y: event.clientY
            };

            // safe current insertion position
            //currentInsertionPos = getInsertionPos(event.target);
            currentInsertionPos = $root.find("g:first").get(0);


            /* check if left mouse button has been pressed (1) or a touch tap occurred (0) */
            if (event.which !== 0 && event.which !== 1) {
                return;
            }

            isMouseDown = true;

            // safe offset
            divOffset = $div.offset();
            if (!divOffset) {
                return;
            }
            divOffset.left -= document.documentElement.scrollLeft;
            divOffset.top -= document.documentElement.scrollTop;
            startPosCursor.x -= divOffset.left;
            startPosCursor.y -= divOffset.top;

            // update selection state
            if (selection.$rect) {
                selection.updateState(event.target);
            }
        }

        function mouseMove(event) {
            var scaleFactor,
                transformedCurrentPosCursor, transformedStartPosCursor,
                deltaX, deltaY;

            event.preventDefault();

            /* check if left mouse button has been pressed (1) or a touch tap occurred (0) */
            if ((event.which !== 0 && event.which !== 1) || !isMouseDown) {
                return;
            }

            disableRectClick = false;

            if (!startPosCursor) {
                return;
            }

            var newX = event.clientX;
            var newY = event.clientY;
            if (divOffset) {
                newX -= divOffset.left;
                newY -= divOffset.top;
            }

            if (selection && (selection.dragging || selection.resizing)) {
                disableRectClick = true;

                //scaleFactor =  1/util.getElementToScreenScaleFactor(width, height, viewBox.width, viewBox.height);
                deltaX = (newX - startPosCursor.x);
                deltaY = (newY - startPosCursor.y);

                if (selection.dragging) {
                    // move rectangle
                    selection.moveRect(deltaX, deltaY);

                } else if (selection.resizing) {
                    // resize rectangle
                    selection.resizeRect(deltaX, deltaY);
                }
            } else {
                // draw new rectangle
                if (popups.metaInfoPopup.isOpen) { // || (popups.navigationPopup && popups.navigationPopup.isOpen)) {
                    return;
                }

                //transformedStartPosCursor = canvasToViewboxCoordinates(startPosCursor.x, startPosCursor.y);
                //transformedCurrentPosCursor = canvasToViewboxCoordinates(event.clientX, event.clientY);

                if (!outlineRect) {
                   outlineRect = addRect({
                       x: 0,
                       y: 0,
                       width: 0,
                       height: 0,
                       parentElement: currentInsertionPos,
                       style: that.OUTLINE_RECT_STYLE
                   });

                    $(outlineRect).on('vmouseup', mouseUp);
                }

                $(outlineRect).attr({
                    x: Math.min(newX, startPosCursor.x),
                    y: Math.min(newY, startPosCursor.y),
                    width: Math.abs(newX - startPosCursor.x),
                    height: Math.abs(newY - startPosCursor.y)
                });
            }
        }

        function mouseUp(event) {
            var rect;
            var dimensions;
            var startPos, currentPos;

            event.preventDefault();

            /* check if left mouse button has been pressed (1) or a touch tap occurred (0) */
            if ((event.which !== 0 && event.which !== 1)) {
                return;
            }

            isMouseDown = false;

            if (selection && (selection.dragging || selection.resizing)) {
                if (selection.dragging) {
                    // rectangle has been moved
                    selection.dragging = false;
                    selection.updatePosition();
                    util.log("rectangle_moved;" + selection.$rect.attr("id"));
                } else if (selection.resizing) {
                    // rectangle has been resized
                    selection.resizing = false;
                    selection.updatePosition();
                    util.log("rectangle_resized;" + selection.$rect.attr("id"));
                }
            } else {
                // draw new rectangle
                if (!startPosCursor || !outlineRect) {
                    return;
                }

                if (outlineRect.width < that.MIN_LENGTH || outlineRect.height < that.MIN_LENGTH) {
                    $(outlineRect).remove();
                    outlineRect = null;
                    return;
                } else {
                    rect = addRect({
                        x: $(outlineRect).attr("x"),
                        y: $(outlineRect).attr("y"),
                        width: $(outlineRect).attr("width"),
                        height: $(outlineRect).attr("height"),
                        parentElement: currentInsertionPos,
                        style: that.RECT_STYLE
                    });

                    $(outlineRect).remove();
                    outlineRect = null;
                }

                //currentPos = canvasToViewboxCoordinates(event.clientX, event.clientY);
                //startPos = canvasToViewboxCoordinates(startPosCursor.x, startPosCursor.y);

//                var newX = event.clientX;
//                var newY = event.clientY;
//                if (divOffset) {
//                    newX -= divOffset.left;
//                    newY -= divOffset.top;
//                }
//
//                dimensions = getRectDimensions(newX, startPosCursor.x, startPosCursor.y, newY);

//                if (dimensions.width < that.MIN_LENGTH || dimensions.height < that.MIN_LENGTH) {
//                    return;
//                }

//                rect = addRect({
//                    x: outlineRect.x,
//                    y: outlineRect.y,
//                    width: outlineRect.width,
//                    height: outlineRect.height,
//                    parentElement: currentInsertionPos,
//                    style: that.RECT_STYLE
//                });

                // disable click on canvas after creation of rect since the panel would be closed otherwise
                disableCanvasClick = true;

                // select new rect after creation
                selection.selectRect(rect);
                popups.panelPopup.open({
                    element: rect
                });

                addToHistory(rect, currentInsertionPos, "add");

                util.log("rectangle_created;" + $(rect).attr("id"));
            }

            startPosCursor = null;

            /*
             * iOS does not trigger click event after changing the DOM during mousemove, therefore reset disableClickForPanel in mouseup event
             * (cp. http://developer.apple.com/library/IOS/#documentation/AppleApplications/Reference/SafariWebContent/HandlingEvents/HandlingEvents.html)
             */
            if (util.isIOSDevice()) {
                disableRectClick = false;
            }
        }

        function getRectDimensions(x1, y1, x2, y2) {
            var left = Math.min(x1, x2);
            var top = Math.min(y1, y2);
            var right = Math.max(x1, x2);
            var bottom = Math.max(y1, y2);

            return {
                x: left,
                y: top,
                width: right - left,
                height: bottom - top
            };
        }

        function click(event) {
            var circle;
            var element, $element;

            if ((event.which !== 0 && event.which !== 1)) {
                return;
            }

            element = event.target;
            $element = $(event.target);

            if ($element.is('rect')) {
                if (!disableRectClick) {
                    //util.log("rectangle_clicked;" + $element.attr("id"));
                    if (popups.panelPopup.isOpen) {
                        if (popups.panelPopup.element == element) { // second click on same rect
                            popups.panelPopup.close();
                        } else { // click on other rect
                            selection.selectRect(element);
                            popups.panelPopup.closeAndReopen(element);
                        }
                    } else { // panel closed
                        selection.selectRect(element);
                        popups.panelPopup.open({
                            element: element
                        });
                    }
                }
            } else if ($element.is('circle')) {
                // do nothing here
            } else if (!disableCanvasClick) {
                // close panel if user clicks on canvas
                if (popups.panelPopup.isOpen) {
                    popups.panelPopup.close();
                }

                selection.clear(false);
            }

            disableRectClick = false;
            disableCanvasClick = false;
        }

        function addRect(spec) {
            var rectElement = util.SVG('rect');

            $(rectElement).attr({
                id: util.generateNewID("m"),
                x: spec.x,
                y: spec.y,
                width: spec.width,
                height: spec.height,
                style: spec.style
            });

            $(spec.parentElement).append(rectElement);

            return rectElement;
        }

        function addCircle(spec) {
            var circleElement = util.SVG('circle');

            $(circleElement).attr({
                id: util.generateNewID("t-"), // temp
                cx: spec.x,
                cy: spec.y,
                r: spec.radius,
                style: spec.style
            });

            $(spec.parentElement).append(circleElement);

            return circleElement;
        }

        function deleteSelectedElement() {
            if (selection.$rect) {
                popups.panelPopup.close();
                addToHistory(selection.$rect, selection.$rect.parent().get(0), "delete");
                //util.log("delete;" + selection.$rect.attr("id"));
                var marker_id = selection.$rect.attr("id");
                window.changelog[marker_id] = {
                    "deleted": true
                }
                selection.clear(true);
            }
        }

        function undo() {
            var historyObject;

            if (history.length === 0 || historyIndex === 0) {
                return;
            }

            selection.clear(false);

            if (popups.panelPopup.isOpen) {
                popups.panelPopup.close();
            }

            historyIndex -= 1;
            historyObject = history[historyIndex];

            switch (historyObject.operation) {
                case "add":
                    $(historyObject.element).detach();
                    break;
                case "delete":
                    $(historyObject.parent).append(historyObject.element);
                    break;
                default:
                    throw new TypeError(APPLICATION_NAME + ": Unknown operation in history.");
            }

            util.log("undo;" + historyObject.operation + ";" + $(historyObject.element).attr("id"));
        }

        function redo() {
            var historyObject;

            if (history.length === 0 || historyIndex == history.length) {
                return;
            }

            if (popups.panelPopup.isOpen) {
                popups.panelPopup.close();
            }

            historyObject = history[historyIndex];

            switch (historyObject.operation) {
                case "add":
                    $(historyObject.parent).append(historyObject.element);
                    break;
                case "delete":
                    $(historyObject.element).detach();
                    break;
                default:
                    throw new TypeError(APPLICATION_NAME + ": Unknown operation in history.");
            }

            historyIndex += 1;

            util.log("redo;" + historyObject.operation + ";" + $(historyObject.element).attr("id"));
        }

        function moveViewBoxLeft() {
            if (viewBox) {
                viewBox.moveLeft();
                util.log("move_left");
            }
        }

        function moveViewBoxRight() {
            if (viewBox) {
                viewBox.moveRight();
                util.log("move_right");
            }
        }

        function moveViewBoxUp() {
            if (viewBox) {
                viewBox.moveUp();
                util.log("move_up");
            }
        }

        function moveViewBoxDown() {
            if (viewBox) {
                viewBox.moveDown();
                util.log("move_down");
            }
        }

        function zoomViewboxIn() {
            if (viewBox) {
                viewBox.zoomIn();
                util.log("zoom_in");
            }
        }

        function zoomViewboxOut() {
            if (viewBox) {
                viewBox.zoomOut();
                util.log("zoom_out");
            }
        }

        function zoomCenter() {
            if (viewBox) {
                viewBox.center(width, height);
                util.log("zoom_center");
            }
        }

        function openMetaInfoPopup() {
            popups.metaInfoPopup.open();
        }

        function setSvgAttributes(attributes) {
            if ($root) {
                $root.attr(attributes);
                //$image.attr("width", attributes.width);
                //$image.attr("height", attributes.height);
            } else {
                throw new ReferenceError(APPLICATION_NAME + ': SVG root is null.');
            }
        }

        function canvasToViewboxCoordinates(xPos, yPos) {
            var scaleFactor,
                newXPos, newYPos;

            if (!viewBox) {
                return null;
            }

            scaleFactor = 1/util.getElementToScreenScaleFactor(width, height, viewBox.width, viewBox.height);
            newXPos = xPos;
            newYPos = yPos;

            if (divOffset) {
                newXPos -= divOffset.left;
                newYPos -= divOffset.top;
            }

            return {
                x: newXPos * scaleFactor + viewBox.x,
                y: newYPos * scaleFactor + viewBox.y
            };
        }

        function getInsertionPos(element) {
            var current = element;

            if ((current == div) || (current == root)) {
                return $('g:first', $root).get(0);
            }

            while (current && !$(current).is('g')) {
                current = $(current).parent().get(0);
            }

            return current;
        }

        function addToHistory(element, parent, operation) {
            history[historyIndex] = {
                element: element,
                parent: parent,
                operation: operation
            };
            historyIndex += 1;
            history.splice(historyIndex); // remove all older nodes after the new node
        }
        //endregion

        //region Privileged methods
        this.initialize = initialize;
        this.undo = undo;
        this.redo = redo;
        this.deleteSelectedElement = deleteSelectedElement;
        this.moveViewBoxUp = moveViewBoxUp;
        this.moveViewBoxDown = moveViewBoxDown;
        this.moveViewBoxLeft = moveViewBoxLeft;
        this.moveViewBoxRight = moveViewBoxRight;
        this.zoomViewboxIn = zoomViewboxIn;
        this.zoomViewboxOut = zoomViewboxOut;
        this.zoomCenter = zoomCenter;
        this.openMetaInfoPopup = openMetaInfoPopup;
        this.addRect = addRect;
        this.addCircle = addCircle;
        this.resize = resize;
        this.uploadSvg = uploadSvg;
        //endregion
    };

    return SvgCanvas;

})(util, PanelPopup, MetaInfoPopup, NavigationPopup, SvgSelection);