/**
 * This component represents the currently selected rectangle and is responsible for moving and resizing it, draws and removes the selection circles.
 *
 * @author baltes
 * @see SvgCanvas.js
 *
 */
var SvgSelection = (function (util) {

    var SvgSelection = function(args) {
        if (!args.svgCanvas) {
            throw new ReferenceError(APPLICATION_NAME + ": SVG canvas is " + args.svgCanvas + ".")
        }

        //region Private members
        var svgCanvas,
            $rect,
            edgeCircles,
            resizingCircle,
            dragging,
            resizing,
            initialXPos,
            initialYPos,
            initialWidth,
            initialHeight;

        svgCanvas = args.svgCanvas;
        //endregion

        //region Getter/setter for private members
        Object.defineProperties(this, {
            "$rect": {
                get: function() {
                    return $rect;
                }
            },
            "selected": {
                get: function() {
                    return ($rect !== null && $rect !== undefined);
                }
            },
            "dragging": {
                get: function() {
                    return dragging;
                },
                set: function(newValue) {
                    dragging = newValue;
                }
            },
            "resizing": {
                get: function() {
                    return resizing;
                },
                set: function(newValue) {
                    resizing = newValue;

                    if (newValue === false) {
                        resizingCircle = null;
                    }
                }
            }
        });
        //endregion

        //region  Private methods
        function selectRect(selectedElement) {
            if (!selectedElement) {
                throw new ReferenceError(APPLICATION_NAME + ": Selected element is " + selectedElement + ".");
            }

            var cssProperties, style,
                radius, circle,
                i;

            clear(false);

            $rect = $(selectedElement);

            updatePosition();

            cssProperties = util.tokenizeStyleString($rect.attr('style'));
            style = svgCanvas.CIRCLE_STYLE;
            style = util.changeStyleProperty(style, "stroke-width", cssProperties["stroke-width"]*0.8);
            style = util.changeStyleProperty(style, "fill", cssProperties["stroke"]);
            radius = (svgCanvas.MIN_LENGTH/2) * 0.9; // circles should not intersect

            edgeCircles = [];

            for (i = 0; i < 4; i++) {
                circle = svgCanvas.addCircle({
                    x: 0,
                    y: 0,
                    radius: radius,
                    parentElement: $rect.parent(),
                    style: style
                });
                edgeCircles.push(circle);
            }

            updateCircles(initialXPos, initialYPos, initialWidth, initialHeight);
        }

        function clear(detachRect) {
            if (detachRect) {
                $rect.detach();
            }

            $(edgeCircles).remove();

            $rect = null;
            edgeCircles = null;
            dragging = false;
            resizing = false;
            resizingCircle = null;
            initialXPos = 0;
            initialYPos = 0;
            initialWidth = 0;
            initialHeight = 0;
        }

        function updateState(clickedElement) {
            var $clickedElement = $(clickedElement);

            if ($clickedElement.is('rect')) {
                // move rectangle
                dragging = true;
            }
            else if ($clickedElement.is('circle')) {
                // resize rectangle
                resizing = true;
                resizingCircle = $clickedElement.get(0);
            }
            else {
                // draw new rectangle
                clear(false);

                if (popups.panelPopup.isOpen) {
                    popups.panelPopup.close();
                }
            }
        }

        function updatePosition() {
            initialXPos = parseFloat($rect.attr('x'));
            initialYPos = parseFloat($rect.attr('y'));
            initialWidth = parseFloat($rect.attr('width'));
            initialHeight = parseFloat($rect.attr('height'));
        }

        function updateCircles(x, y, width, height) {
            if (!edgeCircles) {
                return;
            }

            $(edgeCircles[0]).attr({ cx: x, cy: y });                    //top-left
            $(edgeCircles[1]).attr({ cx: x + width, cy: y });            //top-right
            $(edgeCircles[2]).attr({ cx: x, cy: y + height });           //bottom-left
            $(edgeCircles[3]).attr({ cx: x + width, cy: y + height });   //bottom-right
        }

        function moveRect(deltaX, deltaY) {
            var newX = initialXPos + deltaX;
            var newY = initialYPos + deltaY;
            var width = parseFloat($rect.attr("width"));
            var height = parseFloat($rect.attr("height"));

            $rect.attr({
                x: newX,
                y: newY
            });

            updateCircles(
                newX,
                newY,
                width,
                height
            );
        }

        function resizeRect(deltaX, deltaY) {
            var maxX, maxY;

            maxX = (initialXPos + initialWidth) - svgCanvas.MIN_LENGTH;
            maxY = (initialYPos + initialHeight) - svgCanvas.MIN_LENGTH;

            if (resizingCircle == edgeCircles[0]) { //top-left
                resizeTopLeft(deltaX, deltaY, maxX, maxY);
            }
            else if (resizingCircle == edgeCircles[1]) { //top-right
                resizeTopRight(deltaX, deltaY, maxY);
            }
            else if (resizingCircle == edgeCircles[2]) { //bottom-left
                resizeBottomLeft(deltaX, deltaY, maxX);
            }
            else if (resizingCircle == edgeCircles[3]) { //bottom-right
                resizeBottomRight(deltaX, deltaY);
            }
        }

        function resizeTopLeft(deltaX, deltaY, maxX, maxY) {
            var newX = initialXPos + deltaX;
            var newWidth = initialWidth - deltaX;
            var newY = initialYPos + deltaY;
            var newHeight = initialHeight - deltaY;

            if (newX > maxX) {
                newX = maxX;
                newWidth = svgCanvas.MIN_LENGTH;
            }

            if (newY > maxY) {
                newY = maxY;
                newHeight = svgCanvas.MIN_LENGTH;
            }

            $rect.attr({
                x: newX,
                y: newY,
                width: newWidth,
                height: newHeight
            });

            updateCircles(newX, newY, newWidth, newHeight);
        }

        function resizeTopRight(deltaX, deltaY, maxY) {
            var newX = initialXPos;
            var newY = initialYPos + deltaY;
            var newWidth = initialWidth + deltaX;
            var newHeight = initialHeight - deltaY;

            if (newY > maxY) {
                newY = maxY;
                newHeight = svgCanvas.MIN_LENGTH;
            }

            if (newWidth <= svgCanvas.MIN_LENGTH) {
                newWidth = svgCanvas.MIN_LENGTH;
            }

            if (newHeight <= svgCanvas.MIN_LENGTH) {
                newHeight = svgCanvas.MIN_LENGTH;
            }

            $rect.attr({
                x: newX,
                y: newY,
                width: newWidth,
                height: newHeight
            });

            updateCircles(newX, newY, newWidth, newHeight);
        }

        function resizeBottomLeft(deltaX, deltaY, maxX) {
            var newX = initialXPos + deltaX;
            var newY = initialYPos;
            var newWidth = initialWidth - deltaX;
            var newHeight = initialHeight + deltaY;

            if (newX > maxX) {
                newX = maxX;
                newWidth = svgCanvas.MIN_LENGTH;
            }

            if (newWidth <= svgCanvas.MIN_LENGTH) {
                newWidth = svgCanvas.MIN_LENGTH;
            }

            if (newHeight <= svgCanvas.MIN_LENGTH) {
                newHeight = svgCanvas.MIN_LENGTH;
            }

            $rect.attr({
                x: newX,
                y: newY,
                width: newWidth,
                height: newHeight
            });

            updateCircles(newX, newY, newWidth, newHeight);
        }

        function resizeBottomRight(deltaX, deltaY) {
            var newX = initialXPos;
            var newY = initialYPos;
            var newWidth = initialWidth + deltaX;
            var newHeight = initialHeight + deltaY;

            if (newWidth <= svgCanvas.MIN_LENGTH) {
                newWidth = svgCanvas.MIN_LENGTH;
            }

            if (newHeight <= svgCanvas.MIN_LENGTH) {
                newHeight = svgCanvas.MIN_LENGTH;
            }

            $rect.attr({
                x: newX,
                y: newY,
                width: newWidth,
                height: newHeight
            });

            updateCircles(newX, newY, newWidth, newHeight);
        }
        //endregion

        //region Privileged methods
        this.selectRect = selectRect;
        this.clear = clear;
        this.updateState = updateState;
        this.updatePosition = updatePosition;
        this.updateCircles = updateCircles;
        this.moveRect = moveRect;
        this.resizeRect = resizeRect;
        //endregion
    };

    return SvgSelection;

})(util);