import sketchlink

from sketchlink.exceptions import SketchLinkError


class ProtocolErrorBase(SketchLinkError):
    def __init__(self, corresponding_message, additional_message=""):
        try:
            self.corresponding_message = corresponding_message
            self.error_code = error_codes[self.__class__]
            self.error_metadata = None

            if additional_message:
                self.error_message = error_messages[self.error_code] + ": " + additional_message
            else:
                self.error_message = error_messages[self.error_code] + "."

        except KeyError as e:
            print("Unknown protocol error.")
            raise e

    def __str__(self):
        return str(self.error_code + ": " + self.error_message)


class ArtifactIDNotExistsError(ProtocolErrorBase):
    def __init__(self, corresponding_message, additional_message=""):
        super(ArtifactIDNotExistsError, self).__init__(corresponding_message, additional_message)

        self.error_metadata = {
            "artifactID": corresponding_message.message_data["artifactID"]
        }


class ArtifactIDAlreadyExistsError(ProtocolErrorBase):
    def __init__(self, corresponding_message, additional_message=""):
        super(ArtifactIDAlreadyExistsError, self).__init__(corresponding_message, additional_message)

        self.error_metadata = {
            "artifactID": corresponding_message.message_data["artifactID"]
        }


class ViewerNotAvailableError(ProtocolErrorBase):
    pass


class MessageUnknownError(ProtocolErrorBase):
    pass


class WrongProtocolVersionError(ProtocolErrorBase):
    def __init__(self, corresponding_message, additional_message=""):

        error_message = "expected: " + str(sketchlink.PROTOCOL_VERSION) + ", requested: " + \
                        str(corresponding_message.message_data["protocolVersion"])

        if additional_message:
            error_message += " " + additional_message

        super(WrongProtocolVersionError, self).__init__(corresponding_message, error_message)


class LinkNotExistsError(ProtocolErrorBase):
    def __init__(self, corresponding_message, additional_message=""):
        super(LinkNotExistsError, self).__init__(corresponding_message, additional_message)

        self.error_metadata = {
            "artifactID1": corresponding_message.message_data["artifactID1"],
            "artifactID2": corresponding_message.message_data["artifactID2"]
        }


class ClientAlreadyRegisteredError(ProtocolErrorBase):
    pass


class ClientNotRegisteredError(ProtocolErrorBase):
    pass


class LinkAlreadyExistsError(ProtocolErrorBase):
    def __init__(self, corresponding_message, additional_message=""):
        super(LinkAlreadyExistsError, self).__init__(corresponding_message, additional_message)

        self.error_metadata = {
            "artifactID1": corresponding_message.message_data["artifactID1"],
            "artifactID2": corresponding_message.message_data["artifactID2"]
        }


class InvalidMessageError(ProtocolErrorBase):
    pass


class FieldMissingError(InvalidMessageError):
    @staticmethod
    def from_key_error(message, key_error):
        return FieldMissingError(message, "Field " + str(key_error) + " missing.")


error_codes = {
    ArtifactIDNotExistsError: 1,
    ArtifactIDAlreadyExistsError: 2,
    ViewerNotAvailableError: 3,
    MessageUnknownError: 4,
    WrongProtocolVersionError: 5,
    LinkNotExistsError: 6,
    ClientAlreadyRegisteredError: 7,
    ClientNotRegisteredError: 8,
    LinkAlreadyExistsError: 9,
    InvalidMessageError: 10,
    FieldMissingError: 10
}

error_messages = {
    1: "Artifact ID does not exist",
    2: "Artifact ID already exists",
    3: "Viewer is not available",
    4: "Message unknown",
    5: "Wrong protocol version",
    6: "Link does not exist",
    7: "Client is already registered",
    8: "Client is not registered",
    9: "Link already exists",
    10: "Invalid message",
}