import os
import flask
import re
import sketchlink
import json
import xml.etree.ElementTree as etree

from sketchlink.flask import flask_app
from sketchlink import util, database
from PIL import Image, ExifTags

ALLOWED_EXTENSIONS = set(["jpg", "jpeg"])


@flask_app.route("/sketches/svg/<regex('" + sketchlink.SKETCH_ID_RE + "'):sketch_id>", methods=["GET"])
def get_svg(sketch_id):
    return flask.send_file(util.get_sketch_uri(sketch_id, "svg"))

@flask_app.route("/sketches/svg/<regex('" + sketchlink.SKETCH_ID_RE + "'):sketch_id>", methods=["POST"])
def set_svg(sketch_id):
    request_id = flask.request.form["id"]
    request_svg = flask.request.form["svg"]
    request_changelog = None

    if "changelog" in flask.request.form:
        request_changelog = json.loads(flask.request.form["changelog"])

    if sketch_id != request_id:
        print("Sketch ID and request ID did not match.")
        return ""

    with open(util.get_sketch_uri(sketch_id, "svg"), "w") as svg_file:
        svg_file.write(request_svg) #todo: remove width, height, style attributes from svg root element

    timestamp = util.get_timestamp()

    if request_changelog:

        if sketch_id in request_changelog and util.exists_artifact(sketch_id):
            statement = "UPDATE sketch SET timestamp=? WHERE artifactID=?;"
            database.execute_db(statement, args=(timestamp, sketch_id))

            authors = None
            annotation = None
            title = None
            artifact_id = None

            if "authors" in request_changelog[sketch_id]:
                authors = request_changelog[sketch_id]["authors"]
            if "annotation" in request_changelog[sketch_id]:
                annotation = request_changelog[sketch_id]["annotation"]
            if "title" in request_changelog[sketch_id]:
                title = request_changelog[sketch_id]["title"]
            if "artifact-id" in request_changelog[sketch_id]:
                artifact_id = request_changelog[sketch_id]["artifact-id"]

            if authors:
                statement = "UPDATE sketch SET authors=? WHERE artifactID=?;"
                database.execute_db(statement, args=(authors, sketch_id))

            if annotation:
                statement = "UPDATE sketch SET annotation=? WHERE artifactID=?;"
                database.execute_db(statement, args=(annotation, sketch_id))

            if title:
                statement = "UPDATE sketch SET title=? WHERE artifactID=?;"
                database.execute_db(statement, args=(title, sketch_id))

            if artifact_id:
                if util.exists_artifact(artifact_id) and not util.exists_link(sketch_id, artifact_id):
                    timestamp = util.get_timestamp()
                    statement = "INSERT INTO link VALUES(null, ?, ?, ?, 0);"
                    database.execute_db(statement, args=(sketch_id, artifact_id, timestamp))

        for artifact_id in request_changelog:
            if re.match(sketchlink.MARKER_ID_RE, artifact_id) and "deleted" in request_changelog[artifact_id]:
                statement = "DELETE FROM marker WHERE artifactID=?;"
                database.execute_db(statement, args=(artifact_id,))

        root = etree.fromstring(request_svg)
        g = root[1]
        for marker in g:
            marker_id = marker.attrib.get('id')
            annotation = None
            artifact_id = None

            if marker_id in request_changelog:

                if "annotation" in request_changelog[marker_id]:
                    annotation = request_changelog[marker_id]["annotation"]
                if "artifact-id" in request_changelog[marker_id]:
                    artifact_id = request_changelog[marker_id]["artifact-id"]

                if util.exists_artifact(marker_id):
                    statement = "UPDATE marker SET parentID=?, timestamp=? WHERE artifactID=?;"
                    database.execute_db(statement, args=(sketch_id, timestamp, marker_id))

                    if annotation:
                        statement = "UPDATE marker SET annotation=? WHERE artifactID=?;"
                        database.execute_db(statement, args=(annotation, marker_id))

                else:
                    if not annotation:
                        annotation = ""

                    statement = "INSERT INTO marker VALUES (?, ?, ?, ?);"
                    database.execute_db(statement, args=(marker_id, sketch_id, timestamp, annotation))

                if artifact_id:
                    if util.exists_artifact(artifact_id) and not util.exists_link(marker_id, artifact_id):
                        statement = "INSERT INTO link VALUES(null, ?, ?, ?, 0);"
                        database.execute_db(statement, args=(marker_id, artifact_id, timestamp))

    return ""

@flask_app.route("/sketches/bitmap/<regex('" + sketchlink.SKETCH_ID_RE + "'):sketch_id>", methods=["GET"])
def get_bitmap(sketch_id):
    bitmap_file = util.get_sketch_uri(sketch_id, "jpg")

    if not os.path.exists(bitmap_file):
        bitmap_file = util.get_sketch_uri(sketch_id, "png")

    return flask.send_file(bitmap_file)

@flask_app.route("/sketches/metadata/<regex('" + sketchlink.SKETCH_ID_RE + "'):sketch_id>", methods=["GET"])
def get_sketch_metadata(sketch_id):
    query = "SELECT timestamp, authors, annotation, title FROM sketch WHERE artifactID=?;"
    res = database.query_db(query, args=(sketch_id,), one=True)

    metadata = {}
    if res:
        metadata["timestamp"] = res["timestamp"]
        metadata["authors"] = res["authors"]
        metadata["annotation"] = res["annotation"]
        metadata["title"] = res["title"]

    return flask.Response(json.dumps(metadata), mimetype='application/json')

@flask_app.route("/marker/metadata/<regex('" + sketchlink.MARKER_ID_RE + "'):marker_id>", methods=["GET"])
def get_marker_metadata(marker_id):
    query = "SELECT timestamp, annotation FROM marker WHERE artifactID=?;"
    res = database.query_db(query, args=(marker_id,), one=True)

    metadata = {}
    if res:
        metadata["timestamp"] = res["timestamp"]
        metadata["annotation"] = res["annotation"]

    return flask.Response(json.dumps(metadata), mimetype='application/json')

@flask_app.route("/sketches/links/<regex('" + sketchlink.SKETCH_ID_RE + "'):sketch_id>", methods=["GET"])
def get_sketch_links(sketch_id):
    query = "SELECT artifactID1 as artifactID, deprecated FROM link WHERE artifactID2=? UNION " \
            "SELECT artifactID2 as artifactID, deprecated FROM link WHERE artifactID1=?;"

    res = database.query_db(query, args=(sketch_id, sketch_id))
    linked_artifacts = []

    for row in res:
        linked_artifacts.append(row["artifactID"])

    return flask.Response(json.dumps(linked_artifacts), mimetype='application/json')

@flask_app.route("/marker/links/<regex('" + sketchlink.MARKER_ID_RE + "'):marker_id>", methods=["GET"])
def get_marker_links(marker_id):
    query = "SELECT artifactID1 as artifactID, deprecated FROM link WHERE artifactID2=? UNION " \
            "SELECT artifactID2 as artifactID, deprecated FROM link WHERE artifactID1=?;"

    res = database.query_db(query, args=(marker_id, marker_id))
    linked_artifacts = []

    for row in res:
        linked_artifacts.append(row["artifactID"])

    return flask.Response(json.dumps(linked_artifacts), mimetype='application/json')

@flask_app.route("/sketches/", methods=["POST"])
# flask does not support chunked transfer encoding, which is part of HTTP 1.1.
# Thus, the header field "Content-Length" always has to be set in the request.
def upload_sketch():
    file = flask.request.files["image"]
    if file:
        new_id = util.get_id("s")
        new_folder = os.path.join(sketchlink.SKETCHES_PATH, new_id)
        os.mkdir(new_folder)
        jpg_path = os.path.join(new_folder, new_id + ".jpg")
        svg_path = os.path.join(new_folder, new_id + ".svg")
        file.save(jpg_path)

        jpg_file = Image.open(jpg_path)

        # see http://stackoverflow.com/a/6218425
        #todo: fix bug on Android
        # for orientation in ExifTags.TAGS.keys():
        #     if ExifTags.TAGS[orientation] == 'Orientation':
        #         break
        #
        # if jpg_file._getexif():
        #     exif = dict(jpg_file._getexif().items())
        #     if exif[orientation] == 3:
        #         jpg_file = jpg_file.rotate(180, expand=True)
        #     elif exif[orientation] == 6:
        #         jpg_file = jpg_file.rotate(270, expand=True)
        #     elif exif[orientation] == 8:
        #         jpg_file = jpg_file.rotate(90, expand=True)

        jpg_file.thumbnail((1024, 577), Image.ANTIALIAS)
        jpg_file.save(jpg_path)
        width, height = jpg_file.size

        with open(svg_path, "w") as svg_file:
            svg_file.write(flask.render_template("sketch.svg", id=new_id, width=width, height=height))

        timestamp = util.get_timestamp()
        title = flask.request.form["title"]
        authors = flask.request.form["authors"]
        annotation = flask.request.form["annotation"]

        statement = "INSERT INTO sketch VALUES (?, ?, ?, ?, ?);"
        database.execute_db(statement, args=(new_id, timestamp, authors, annotation, title))

        artifact_id = flask.request.form["artifact-id"]
        if artifact_id and util.exists_artifact(artifact_id) and not util.exists_link(new_id, artifact_id):
            timestamp = util.get_timestamp()
            statement = "INSERT INTO link VALUES(null, ?, ?, ?, 0);"
            database.execute_db(statement, args=(new_id, artifact_id, timestamp))

        return new_id

@flask_app.route("/sketches/", methods=["GET"])
def get_sketches():
    query = "SELECT artifactID FROM sketch"
    res = database.query_db(query)
    sketches = []
    for row in res:
        sketches.append(row["artifactID"])

    return flask.Response(json.dumps(sketches), mimetype='application/json')