import copy
import uuid
import time
from random import randint

from tornado import escape

import sketchlink
import sketchlink.tornado
from sketchlink import util
from sketchlink.exceptions import general_errors, protocol_errors
from sketchlink.database import query_db, execute_db
from sketchlink.util import exists_artifact


class Message(object):
    def __init__(self, message_id=str(uuid.uuid4()), message_name="unknown", message_type="unknown", message_data=None):
        self.message_id = message_id
        self.message_name = message_name
        self.message_type = message_type
        self.message_data = message_data

    @staticmethod
    def from_json(json_string):
        # default values
        message_dict = None
        message_id = str(uuid.uuid4())
        message_name = "unknown"
        message_type = "unknown"
        message_data = None

        try:
            message_dict = escape.json_decode(json_string)
        except (TypeError, ValueError) as e:
            raise protocol_errors.InvalidMessageError(Message(), "Error while parsing JSON: " + str(e))

        try:
            message_id = message_dict["id"]
            message_name = message_dict["message"]
            message_type = message_dict["type"]
            message_data = message_dict["data"]
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(
                Message(message_id, message_name, message_type, message_data),
                e
            )

        if not util.validate_message_id(message_id):
            raise protocol_errors.InvalidMessageError(
                Message(message_id, message_name, message_type, message_data),
                "Invalid ID " + message_id
            )

        if message_type == "request":
            return Request(message_id, message_name, message_data)
        elif message_type == "response":
            return Response(message_id, message_name, message_data)
        elif message_type == "error":
            return Error(message_id, message_name, message_data)
        elif message_type == "broadcast":
            return Broadcast(message_id, message_name, message_data)
        else:
            raise protocol_errors.InvalidMessageError(
                Message(message_id, message_name, message_type, message_data),
                "Invalid type " + message_type
            )

    def to_json(self):
        json_dict = {
            "id": self.message_id,
            "message": self.message_name,
            "type": self.message_type,
            "data": self.message_data
        }

        return escape.json_encode(json_dict)

    def __str__(self):
        return str(self.to_json())


class Request(Message):
    def __init__(self, message_id=str(uuid.uuid4()), message_name="", message_data=None):
        super(Request, self).__init__(message_id, message_name, "request", message_data)


class Error(Message):
    def __init__(self, message_id=str(uuid.uuid4()), message_name="", message_data=None):
        super(Error, self).__init__(message_id, message_name, "error", message_data)

    @staticmethod
    def from_exception(exception):
        if not isinstance(exception, protocol_errors.ProtocolErrorBase):
            raise general_errors.InvalidArgumentError("exception", exception)
        return Error(
            exception.corresponding_message.message_id,
            exception.corresponding_message.message_name,
            {
                "code": exception.error_code,
                "message": exception.error_message,
                "metaData": exception.error_metadata
            }
        )


class Response(Message):
    def __init__(self, message_id=str(uuid.uuid4()), message_name="", message_data=None):
        super(Response, self).__init__(message_id, message_name, "response", message_data)

    @staticmethod
    def from_request(request, client, connected_clients, display_names):
        if not isinstance(request, Request):
            raise general_errors.InvalidArgumentError("request_message", request.__class__)

        if request.message_name == "register":
            return Response._from_register(request, client, display_names)
        elif request.message_name == "getAvailableClients":
            return Response._from_get_available_clients(request, connected_clients)
        elif request.message_name == "getLinkedArtifacts":
            return Response._from_get_linked_artifacts(request)
        elif request.message_name == "getContainedAnchors":
            return Response._from_get_contained_anchors(request)
        elif request.message_name == "getArtifactMetaData":
            return Response._from_get_artifact_metadata(request)
        elif request.message_name == "exists":
            return Response._from_exists(request)
        elif request.message_name == "createArtifact":
            return Response._from_create_artifact(request)
        elif request.message_name == "updateArtifactMetaData":
            return Response._from_update_artifact(request)
        elif request.message_name == "createLink":
            return Response._from_create_link(request)
        elif request.message_name == "updateLink":
            return Response._from_update_link(request)
        elif request.message_name == "deleteLink":
            return Response._from_delete_link(request)
        elif request.message_name == "log":
            return Response._from_log(request, client)
        elif request.message_name == "getSketches":
            return Response._from_get_sketches(request)
        else:
            raise protocol_errors.MessageUnknownError(
                request,
                "Message name was " + request.message_name
            )

    @staticmethod
    def _from_register(request, client, display_names):
        try:
            protocol_version = request.message_data["protocolVersion"]
            preferred_name = request.message_data["preferredDisplayName"]
            client_type = request.message_data["clientType"]
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        if protocol_version == sketchlink.PROTOCOL_VERSION:
            if client.is_registered:
                raise protocol_errors.ClientAlreadyRegisteredError(request)
            else:
                if preferred_name in display_names:
                    preferred_name += str(randint(1, 9)) + str(randint(0, 9)) + str(randint(0, 9))

                client.display_name = preferred_name
                client.is_registered = True
                client.client_type = client_type

                display_names.append(preferred_name)

                return Message(request.message_id, request.message_name, "response", {
                    "displayName": preferred_name,
                    "protocolVersion": sketchlink.PROTOCOL_VERSION,
                    "identifier": client.identifier
                })
        else:
            raise protocol_errors.WrongProtocolVersionError(request)

    @staticmethod
    def _from_get_available_clients(request, connected_clients):
        available_clients = []

        for connected_client in connected_clients:
            available_clients.append({
                "displayName": connected_client.display_name,
                "identifier": connected_client.identifier,
                "type": connected_client.client_type,
                "ipAddress": connected_client.ip_address
            })

        return Response(
            request.message_id,
            request.message_name,
            {
                "clients": available_clients
            }
        )

    @staticmethod
    def _from_get_linked_artifacts(request):
        try:
            artifact_id = str(request.message_data["artifactID"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        if not exists_artifact(artifact_id):
            raise protocol_errors.ArtifactIDNotExistsError(request)

        query = "SELECT artifactID1 as artifactID, deprecated FROM link WHERE artifactID2=? UNION " \
                "SELECT artifactID2 as artifactID, deprecated FROM link WHERE artifactID1=?;"

        res = query_db(query, args=(artifact_id, artifact_id))
        linked_artifacts = []

        for row in res:
            linked_artifact_id = row["artifactID"]

            if linked_artifact_id.startswith('m'):
                tmp = query_db("SELECT parentID FROM marker WHERE artifactID=?", args=(linked_artifact_id,), one=True)
                parent_id = tmp["parentID"]
            elif linked_artifact_id.startswith('c'):
                tmp = query_db("SELECT parentID FROM anchor WHERE artifactID=?", args=(linked_artifact_id,), one=True)
                parent_id = tmp["parentID"]
            else:
                parent_id = None

            linked_artifacts.append({
                "artifactID": linked_artifact_id,
                "parentID": parent_id,
                "deprecated": bool(row["deprecated"])
            })

        return Response(request.message_id, request.message_name, {
            "artifactID": artifact_id,
            "linkedArtifacts": linked_artifacts
        })

    @staticmethod
    def _from_get_contained_anchors(request):
        try:
            path = str(request.message_data["path"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        res = query_db("SELECT artifactID FROM anchor WHERE path=?;", args=(path,))
        contained_anchors = []

        for row in res:
            contained_anchors.append(row["artifactID"])

        return Response(request.message_id, request.message_name, {
            "path": path,
            "containedAnchors": contained_anchors
        })

    @staticmethod
    def _from_get_artifact_metadata(request):
        try:
            artifact_id = str(request.message_data["artifactID"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        if artifact_id.startswith("s"):
            query = "SELECT timestamp, authors, annotation, title FROM sketch WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if not res:
                raise protocol_errors.ArtifactIDNotExistsError(request)

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id,
                "metaData": {
                    "timestamp": res["timestamp"],
                    "authors": res["authors"],
                    "annotation": res["annotation"],
                    "title": res["title"]
                }
            })

        elif artifact_id.startswith("m"):
            query = "SELECT parentID, timestamp, annotation FROM marker WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if not res:
                raise protocol_errors.ArtifactIDNotExistsError(request)

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id,
                "metaData": {
                    "parentID": res["parentID"],
                    "timestamp": res["timestamp"],
                    "annotation": res["annotation"]
                }
            })

        elif artifact_id.startswith("c"):
            query = "SELECT parentID, timestamp, path, targetType, targetName, projectName FROM anchor WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if not res:
                raise protocol_errors.ArtifactIDNotExistsError(request)

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id,
                "metaData": {
                    "parentID": res["parentID"],
                    "timestamp": res["timestamp"],
                    "path": res["path"],
                    "targetType": res["targetType"],
                    "targetName": res["targetName"],
                    "projectName": res["projectName"]
                }
            })

        else:
            raise protocol_errors.ArtifactIDNotExistsError(request)

    @staticmethod
    def _from_exists(request):
        try:
            artifact_id = str(request.message_data["artifactID"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        return Response(request.message_id, request.message_name, {
            "artifactID": artifact_id,
            "exists": util.exists_artifact(artifact_id)
        })

    @staticmethod
    def _from_create_artifact(request):
        try:
            artifact_id = str(request.message_data["artifactID"])
            meta_data = request.message_data["metaData"]
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        if artifact_id.startswith("s"):
            query = "SELECT artifactID FROM sketch WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if res:
                raise protocol_errors.ArtifactIDAlreadyExistsError(request)

            try:
                authors = meta_data["authors"]
                annotation = meta_data["annotation"]
                title = meta_data["title"]
            except KeyError as e:
                raise protocol_errors.FieldMissingError.from_key_error(request, e)

            timestamp = util.get_timestamp()
            statement = "INSERT INTO sketch VALUES (?, ?, ?, ?, ?);"
            execute_db(statement, args=(artifact_id, timestamp, authors, annotation, title))

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id
            })

        elif artifact_id.startswith("m"):
            query = "SELECT artifactID FROM marker WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if res:
                raise protocol_errors.ArtifactIDAlreadyExistsError(request)

            try:
                parent_id = meta_data["parentID"]
                timestamp = util.get_timestamp()
                annotation = meta_data["annotation"]
            except KeyError as e:
                raise protocol_errors.FieldMissingError.from_key_error(request, e)

            statement = "INSERT INTO marker VALUES (?, ?, ?, ?);"
            execute_db(statement, args=(artifact_id, parent_id, timestamp, annotation))

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id
            })

        elif artifact_id.startswith("c"):
            query = "SELECT artifactID FROM anchor WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if res:
                raise protocol_errors.ArtifactIDAlreadyExistsError(request)

            try:
                parent_id = meta_data["parentID"]
                timestamp = util.get_timestamp()
                path = meta_data["path"]
                target_type = meta_data["targetType"]
                target_name = meta_data["targetName"]
                project_name = meta_data["projectName"]
            except KeyError as e:
                raise protocol_errors.FieldMissingError.from_key_error(request, e)

            statement = "INSERT INTO anchor VALUES (?, ?, ?, ?, ?, ?, ?);"
            execute_db(statement, args=(artifact_id, parent_id, timestamp, path, target_type, target_name, project_name))

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id
            })

        else:
            raise protocol_errors.ArtifactIDAlreadyExistsError(request)

    @staticmethod
    def _from_update_artifact(request):
        try:
            artifact_id = str(request.message_data["artifactID"])
            meta_data = request.message_data["metaData"]
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        timestamp = util.get_timestamp()

        if artifact_id.startswith("s"):
            query = "SELECT artifactID FROM sketch WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if not res:
                raise protocol_errors.ArtifactIDNotExistsError(request)

            try:
                authors = meta_data["authors"]
                annotation = meta_data["annotation"]
                title = meta_data["title"]
            except KeyError as e:
                raise protocol_errors.FieldMissingError.from_key_error(request, e)

            statement = "UPDATE sketch SET timestamp=?, authors=?, annotation=?, title=? WHERE artifactID=?;"
            execute_db(statement, args=(timestamp, authors, annotation, title, artifact_id))

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id
            })

        elif artifact_id.startswith("m"):
            query = "SELECT artifactID FROM marker WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if not res:
                raise protocol_errors.ArtifactIDNotExistsError(request)

            try:
                parent_id = meta_data["parentID"]
                annotation = meta_data["annotation"]
            except KeyError as e:
                raise protocol_errors.FieldMissingError.from_key_error(request, e)

            statement = "UPDATE marker SET parentID=?, timestamp=?, annotation=? WHERE artifactID=?;"
            execute_db(statement, args=(parent_id, timestamp, annotation, artifact_id))

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id
            })

        elif artifact_id.startswith("c"):
            query = "SELECT artifactID FROM anchor WHERE artifactID=?;"
            res = query_db(query, args=(artifact_id,), one=True)

            if not res:
                raise protocol_errors.ArtifactIDNotExistsError(request)

            try:
                parent_id = meta_data["parentID"]
                path = meta_data["path"]
                target_type = meta_data["targetType"]
                target_name = meta_data["targetName"]
                project_name = meta_data["projectName"]
            except KeyError as e:
                raise protocol_errors.FieldMissingError.from_key_error(request, e)

            statement = "UPDATE anchor SET parentID=?, timestamp=?, path=?, targetType=?, targetName=?, projectName=? " \
                        "WHERE artifactID=?;"
            execute_db(statement, args=(parent_id, timestamp, path, target_type, target_name, project_name, artifact_id))

            return Response(request.message_id, request.message_name, {
                "artifactID": artifact_id
            })
        else:
            raise protocol_errors.ArtifactIDNotExistsError(request)

    @staticmethod
    def _from_create_link(request):
        try:
            artifact_id_1 = str(request.message_data["artifactID1"])
            artifact_id_2 = str(request.message_data["artifactID2"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        if not util.exists_artifact(artifact_id_1):
            request.message_data["artifactID"] = artifact_id_1
            raise protocol_errors.ArtifactIDNotExistsError(request)

        elif not util.exists_artifact(artifact_id_2):
            request.message_data["artifactID"] = artifact_id_2
            raise protocol_errors.ArtifactIDNotExistsError(request)

        elif util.exists_link(artifact_id_1, artifact_id_2):
            raise protocol_errors.LinkAlreadyExistsError(request)

        else:
            timestamp = util.get_timestamp()
            statement = "INSERT INTO link VALUES(null, ?, ?, ?, 0);"
            execute_db(statement, args=(artifact_id_1, artifact_id_2, timestamp))

            return Response(request.message_id, request.message_name, {
                "artifactID1": artifact_id_1,
                "artifactID2": artifact_id_2
            })

    @staticmethod
    def _from_update_link(request):
        try:
            artifact_id_1 = str(request.message_data["artifactID1"])
            artifact_id_2 = str(request.message_data["artifactID2"])
            deprecated = bool(request.message_data["deprecated"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        if util.exists_link(artifact_id_1, artifact_id_2):
            deprecated = (1 if deprecated else 0)  # sqlite has no boolean type
            timestamp = util.get_timestamp()
            statement = "UPDATE link SET deprecated=?, timestamp=? WHERE (artifactID1=? AND artifactID2=?) OR " \
                        "(artifactID1=? AND artifactID2=?);"
            execute_db(statement, args=(deprecated, timestamp, artifact_id_1, artifact_id_2, artifact_id_2, artifact_id_1))

            return Response(request.message_id, request.message_name, {
                "artifactID1": artifact_id_1,
                "artifactID2": artifact_id_2,
                "deprecated": (True if deprecated else False)
            })
        else:
            raise protocol_errors.LinkNotExistsError(request)

    @staticmethod
    def _from_delete_link(request):
        try:
            artifact_id_1 = str(request.message_data["artifactID1"])
            artifact_id_2 = str(request.message_data["artifactID2"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        query = "SELECT artifactID1, artifactID2, deprecated FROM link WHERE (artifactID1=? AND artifactID2=?) " \
                "OR (artifactID1=? AND artifactID2=?);"
        res = query_db(query, args=(artifact_id_1, artifact_id_2, artifact_id_2, artifact_id_1), one="True")

        if res:
            statement = "DELETE FROM link WHERE (artifactID1=? AND artifactID2=?) OR " \
                        "(artifactID1=? AND artifactID2=?);"
            execute_db(statement, args=(artifact_id_1, artifact_id_2, artifact_id_2, artifact_id_1))

            timestamp = util.get_timestamp()
            deprecated = res["deprecated"]
            statement = "INSERT INTO linkArchive VALUES (null, ?, ?, ?, ?);"
            execute_db(statement, args=(artifact_id_1, artifact_id_2, timestamp, deprecated))

            return Response(request.message_id, request.message_name, {
                "artifactID1": None,
                "artifactID2": None,
            })
        else:
            raise protocol_errors.LinkNotExistsError(request)

    @staticmethod
    def _from_log(request, client):
        try:
            message = str(request.message_data["message"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        millis = int(round(time.time() * 1000))

        sketchlink.tornado.log_file.write(";".join([str(millis), message,
                                                    client.display_name, client.identifier, client.ip_address,
                                                    client.client_type]) + "\n")

        return None

    @staticmethod
    def _from_get_sketches(request):
        try:
            filter_string = str(request.message_data["filter"])
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(request, e)

        sketches = []
        query = ""

        if filter_string == "all":
            query = "SELECT DISTINCT artifactID FROM sketch " \
                    "ORDER BY timestamp DESC;"

        elif filter_string == "unlinked":
            query = "SELECT artifactID, timestamp FROM sketch WHERE artifactID NOT IN " \
                    "(" \
                    "SELECT artifactID FROM sketch AS S JOIN link AS L ON " \
                    "(S.artifactID=L.artifactID1 OR S.artifactID=L.artifactID2) WHERE L.deprecated<>1 " \
                    "UNION " \
                    "SELECT distinct parentID from marker as M JOIN link as L on " \
                    "(M.artifactID=L.artifactID1 OR M.artifactID=L.artifactID2) WHERE L.deprecated<>1 " \
                    ")" \
                    "ORDER BY timestamp DESC;"

        res = query_db(query)

        for row in res:
            sketches.append(row[0])

        return Response(request.message_id, request.message_name, {
            "filter": filter_string,
            "sketches": sketches
        })


class Broadcast(Message):
    def __init__(self, message_id=str(uuid.uuid4()), message_name="", message_data=None):
        super(Broadcast, self).__init__(message_id, message_name, "broadcast", message_data)

    @staticmethod
    def from_broadcast(message, client, connected_clients):
        try:
            received_identifier_list = message.message_data["clients"]
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(message, e)

        if message.message_name == "show":
            Broadcast._validate_show(message)
        else:
            raise protocol_errors.MessageUnknownError(message)

        broadcast_identifier_list = []
        broadcast_clients = []

        if not received_identifier_list:  # broadcast to all clients
            broadcast_clients = connected_clients
            for current_client in connected_clients:
                broadcast_identifier_list.append(current_client.identifier)
        else:
            received_identifier_list = list(received_identifier_list)
            for current_client in filter(lambda x: x.identifier in received_identifier_list, connected_clients):
                broadcast_clients.append(current_client)
                broadcast_identifier_list.append(current_client.identifier)

        broadcast = copy.deepcopy(message)
        broadcast.message_data["clients"] = broadcast_identifier_list
        broadcast.message_data["origin"] = client.identifier

        return broadcast, broadcast_clients

    @staticmethod
    def _validate_show(message):
        try:
            message.message_data["artifactID"]
        except KeyError as e:
            raise protocol_errors.FieldMissingError.from_key_error(message, e)
